#include "DXUT.h"
#include "App.h"
#include <assimp.hpp>      // C++ importer interface
#include <aiScene.h>       // Output data structure
#include <aiPostProcess.h> // Post processing flags

App::App(LPDIRECT3DDEVICE9 device)
	:EngineInterface(device)
{
	camera.SetViewParams(
		&D3DXVECTOR3(10, 10, 10),
		&D3DXVECTOR3(0, 0, 0));
}

HRESULT App::createManagedResources()
{
	LPD3DXBUFFER compilationErrors;
	if(FAILED(
		D3DXCreateEffectFromFile(device, L"envmap.fx", NULL, NULL, 0, NULL, &effect, &compilationErrors))) 
	{
		if(compilationErrors)
			MessageBoxA( NULL,
			(LPSTR)compilationErrors->GetBufferPointer(),
			"Failed to load effect file!", MB_OK);
		exit(-1);
	}

	LPD3DXBUFFER materialBuffer;
	D3DXLoadMeshFromX(L"media\\testtorus.x", D3DXMESH_MANAGED, device, NULL,  &materialBuffer, NULL, (DWORD*)&nSubMeshes, &mesh);


	D3DXMATERIAL* materialArray = (D3DXMATERIAL*)
	materialBuffer->GetBufferPointer();

	for(int t=0; t<nSubMeshes; t++)
	{
		LPDIRECT3DTEXTURE9 texture;
		char textureFilePath[512];
		strcpy(textureFilePath, "media\\");
		strcat(textureFilePath, 	materialArray[t].pTextureFilename);
		HRESULT hr = D3DXCreateTextureFromFileA(
			device, textureFilePath, &texture);
		if(hr == S_OK)
			textures.push_back(texture);
		else
			textures.push_back(NULL);      
	}

	D3DXCreateCubeTextureFromFile(device, L"media\\cloudyNoon.dds", &environmentCube);

	return S_OK; 
}

HRESULT App::releaseManagedResources()
{
	environmentCube->Release();

	std::vector<LPDIRECT3DTEXTURE9>::iterator 	i = textures.begin();
	while(i != textures.end())
	{
		if(*i != NULL)
			(*i)->Release(); i++;
	}
	mesh->Release();
	effect->Release();
	return S_OK;
}

HRESULT App::createDefaultResources()
{
	device->CreateVertexBuffer(
		4 * sizeof(QuadVertex), D3DUSAGE_DYNAMIC|D3DUSAGE_WRITEONLY, D3DFVF_XYZ |  D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE3(0),
		D3DPOOL_DEFAULT,
		&quadVertexBuffer, NULL);

	QuadVertex* vertexData;
	quadVertexBuffer->Lock(0, 0, 	(void**)&vertexData, D3DLOCK_DISCARD);
		vertexData[0].pos = D3DXVECTOR3(-1.0f, 1.0f, 0.0f);
		vertexData[1].pos = D3DXVECTOR3(1.0f, 1.0f, 0.0f);
		vertexData[2].pos = D3DXVECTOR3(-1.0f, -1.0f, 0.0f);
		vertexData[3].pos = D3DXVECTOR3(1.0f, -1.0f, 0.0f);

		vertexData[0].tex = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
		vertexData[1].tex = D3DXVECTOR3(1.0f, 0.0f, 0.0f);
		vertexData[2].tex = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
		vertexData[3].tex = D3DXVECTOR3(1.0f, 1.0f, 0.0f);
	quadVertexBuffer->Unlock();


	const D3DSURFACE_DESC* desc = DXUTGetD3D9BackBufferSurfaceDesc();
	camera.SetProjParams(3.14f / 2.0f, (float)desc->Width / desc->Height, 0.1, 1000.0);
	return S_OK;
}

HRESULT App::releaseDefaultResources()
{
	quadVertexBuffer->Release();
	return S_OK;
}

void App::processMessage(
	HWND hWnd, UINT uMsg,
	WPARAM wParam, LPARAM lParam)
{
	if(uMsg == WM_KEYUP)
	{
		if(wParam == VK_F3)
			openFile();
		if(wParam == VK_F2)
			saveFile();
	}

	camera.HandleMessages(
			hWnd, uMsg, wParam, lParam);
}

void App::animate(double dt, double t)
{
	camera.FrameMove(dt);
}


void App::render()
{
	HRESULT hr;

	device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(0, 45, 50, 170), 1.0f, 0);

	D3DXMATRIX modelMatrix;
	D3DXMATRIX modelMatrixInverse;
	D3DXMATRIX modelViewProjMatrix;

	D3DXMatrixIdentity(&modelMatrix);
	D3DXMatrixInverse(&modelMatrixInverse, NULL, &modelMatrix);
	modelViewProjMatrix =
		modelMatrix * (*camera.GetViewMatrix()) * (*camera.GetProjMatrix());

	effect->SetMatrix( effect->GetParameterByName(0, "modelMatrix"), 	&modelMatrix);
	effect->SetMatrix(effect->GetParameterByName(0, "modelMatrixInverse"), 	&modelMatrixInverse);
	effect->SetMatrix(effect->GetParameterByName(0, "modelViewProjMatrix"), 	&modelViewProjMatrix);

	const D3DXVECTOR3& eyePos = *camera.GetEyePt();
	D3DXMATRIX eyePosTranslationMatrix;
	D3DXMatrixTranslation(&eyePosTranslationMatrix, eyePos.x, eyePos.y, eyePos.z);
	D3DXMATRIX rpmi;
	D3DXMatrixInverse(&rpmi, NULL, &(eyePosTranslationMatrix * *camera.GetViewMatrix() * *camera.GetProjMatrix()));
	effect->SetMatrix(effect->GetParameterByName(0, 
	"rotProjMatrixInverse"),
		 &rpmi);

	effect->
		SetTexture(effect->
		GetParameterByName(0, "environmentCubeTexture"),
		environmentCube);


	effect->CommitChanges();

	UINT a;

	device->BeginScene();
		effect->SetTechnique(effect->
			GetTechniqueByName("environment"));
		effect->Begin(&a, 0);
		effect->BeginPass(0);

			device->SetStreamSource(0,
				quadVertexBuffer, 0,
				sizeof(QuadVertex));
			device->SetFVF(D3DFVF_XYZ |  D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE3(0));
			device->DrawPrimitive
				(D3DPT_TRIANGLESTRIP, 0, 2);
		
		effect->EndPass();
		effect->End();


		effect->SetTechnique( effect->GetTechniqueByName("show") );
		effect->Begin(&a, 0);
		effect->BeginPass(0);

			for(int i=0; i<nSubMeshes; i++)
			{
				effect->SetTexture(
					effect->GetParameterByName(0, "kdMap"), textures.at(i));
				effect->CommitChanges();
				mesh->DrawSubset(i);
			}

		effect->EndPass();
		effect->End();

	device->EndScene();
}


void App::openFile()
{
	char filename[512];
	char ext[512];
	filename[0] = L'\0';
	OPENFILENAMEA of;
	of.lStructSize = sizeof(OPENFILENAME);
	of.hwndOwner = DXUTGetHWND();
	of.hInstance = NULL;
	of.lpstrFilter = "Collada (*.dae, *.xml)\0*.dae;*.xml\0Biovision BVH (*.bvh)\0*.bvh\0 3D Studio Max 3DS (*.3ds)\0*.3ds\0 3D Studio Max ASE (*.ase)\0*.ase\0Wavefront Object (*.obj)\0*.obj\0Stanford Polygon Library (*.ply)\0*.ply\0AutoCAD DXF (*.dxf)\0*.dxf\0Neutral File Format (*.nff)\0*.nff\0Sense8 WorldToolkit (*.nff)\0*.nff\0Valve Model (*.smd,*.vta)\0*.smd,*.vta\0Quake I (*.mdl)\0*.mdl\0Quake II (*.md2)\0*.md2\0Quake III (*.md3)\0*.md3\0RtCW (*.mdc)\0*.mdc\0Doom 3 (*.md5mesh;*.md5anim;*.md5camera)\0*.md5mesh;*.md5anim;*.md5camera\0DirectX X (*.x)\0*.x\0Quick3D (*.q3o, *.q3s)\0*.q3o;*q3s\0Raw Triangles (*.raw)\0*.raw\0AC3D (*.ac)\0*.ac\0Stereolithography (*.stl)\0*.stl\0Autodesk DXF (*.dxf)\0*.dxf\0Irrlicht Mesh (*.irrmesh, *.xml)\0*.irrmesh;*.xml\0Irrlicht Scene (*.irr, *.xml)\0*.irr;*.xml\0Object File Format (*.off)\0*.off\0Terragen Terrain (*.ter)\0*.ter\0 3D GameStudio Model (*.mdl)\0*.mdl\0 3D GameStudio Terrain (*.hmp)\0*.hmp\0Ogre (*.mesh.xml, *.skeleton.xml, *.material)\0*.mesh.xml;*.skeleton.xml;*.material\0Milkshape 3D (*.ms3d)\0*.ms3d\0LightWave Model (*.lwo)\0*.lwo\0LightWave Scene (*.lws)\0*.lws\0Modo Model (*.lxo)\0*.lxo\0CharacterStudio Motion (*.csm)\0*.csm\0Stanford Ply (*.ply)\0*.ply\0TrueSpace (*.cob, *.scn)\0*.cob;*.scn\0\0";
	of.lpstrCustomFilter = NULL;
	of.nMaxCustFilter = 0;
	of.nFilterIndex = 0;
	of.lpstrFile = filename;
	of.nMaxFile = 512; 
	of.lpstrFileTitle = NULL;
	of.nMaxFileTitle = 0;
	of.lpstrInitialDir = ".";
	of.lpstrTitle = NULL;
	of.Flags = 0; 
	of.nFileOffset = 0; 
	of.nFileExtension = 0; 
	of.lpstrDefExt = NULL;
	of.lCustData = NULL;
	of.lpfnHook = NULL;
	of.lpTemplateName = NULL; 
	of.pvReserved = NULL;
	of.dwReserved = NULL;
	of.FlagsEx = 0;

	GetOpenFileNameA(&of);

	if(filename[0] == '\0')
		return;

	Assimp::Importer importer;

	const aiScene* assScene = importer.ReadFile( filename,
	//	aiProcess_FlipWindingOrder		|
		aiProcess_CalcTangentSpace		| 
		aiProcess_Triangulate			|
		aiProcess_JoinIdenticalVertices	|
		aiProcess_SortByPType);
  
	// if the import failed
	if( !assScene || !assScene->HasMeshes() || assScene->mNumMeshes == 0) 
	{
		mesh = NULL;
		MessageBoxA( NULL, 
			"Mesh could not be loaded.",
			"Failed to create mesh from file!", MB_OK);
		exit(-1);
	}

	nSubMeshes = assScene->mNumMeshes;

	// count vertices and faces
	unsigned int cVertices = 0;
	unsigned int cFaces = 0;
	D3DXATTRIBUTERANGE* submeshRanges = new D3DXATTRIBUTERANGE[assScene->mNumMeshes];
	for(unsigned int iAssMesh=0; iAssMesh< assScene->mNumMeshes; iAssMesh++)
	{
		submeshRanges[iAssMesh].AttribId = iAssMesh;
		submeshRanges[iAssMesh].FaceCount = assScene->mMeshes[iAssMesh]->mNumFaces;
		submeshRanges[iAssMesh].FaceStart = cFaces;
		submeshRanges[iAssMesh].VertexCount = assScene->mMeshes[iAssMesh]->mNumVertices;
		submeshRanges[iAssMesh].VertexStart = cVertices;

		cFaces += assScene->mMeshes[iAssMesh]->mNumFaces;
		cVertices += assScene->mMeshes[iAssMesh]->mNumVertices;
	}

	bool wideIndexBuffer = (cFaces > 0xffff);

	LPD3DXMESH wMesh;
	D3DXCreateMeshFVF( cFaces, cVertices, wideIndexBuffer?D3DXMESH_32BIT:0, D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE3(0), device, &wMesh );
	//wMesh->SetAttributeTable(submeshRanges, assScene->mNumMeshes);

	DWORD* attribData;
	unsigned int iAttrib=0;
	wMesh->LockAttributeBuffer(0, &attribData);
	for(unsigned int iAssMesh=0; iAssMesh< assScene->mNumMeshes; iAssMesh++)
	{
		for(unsigned int iAssFace=0; iAssFace< assScene->mMeshes[iAssMesh]->mNumFaces; iAssFace++)
		{
			attribData[iAttrib++] =  iAssMesh;
		}
	}
	wMesh->UnlockAttributeBuffer();

	if(wideIndexBuffer)
	{
		unsigned int* indexData;
		wMesh->LockIndexBuffer(0, (void**)&indexData);
		unsigned int iIndex=0;
		for(unsigned int iAssMesh=0; iAssMesh< assScene->mNumMeshes; iAssMesh++)
		{
			for(unsigned int iAssFace=0; iAssFace< assScene->mMeshes[iAssMesh]->mNumFaces; iAssFace++)
			{
				indexData[iIndex++] =  assScene->mMeshes[iAssMesh]->mFaces[iAssFace].mIndices[0] + submeshRanges[iAssMesh].VertexStart;
				indexData[iIndex++] =  assScene->mMeshes[iAssMesh]->mFaces[iAssFace].mIndices[1] + submeshRanges[iAssMesh].VertexStart;
				indexData[iIndex++] =  assScene->mMeshes[iAssMesh]->mFaces[iAssFace].mIndices[2] + submeshRanges[iAssMesh].VertexStart;
			}
		}
	}
	else
	{
		unsigned short* indexData;
		wMesh->LockIndexBuffer(0, (void**)&indexData);
		unsigned int iIndex=0;
		for(unsigned int iAssMesh=0; iAssMesh< assScene->mNumMeshes; iAssMesh++)
		{
			for(unsigned int iAssFace=0; iAssFace< assScene->mMeshes[iAssMesh]->mNumFaces; iAssFace++)
			{
				indexData[iIndex++] =  assScene->mMeshes[iAssMesh]->mFaces[iAssFace].mIndices[0] + submeshRanges[iAssMesh].VertexStart;
				indexData[iIndex++] =  assScene->mMeshes[iAssMesh]->mFaces[iAssFace].mIndices[1] + submeshRanges[iAssMesh].VertexStart;
				indexData[iIndex++] =  assScene->mMeshes[iAssMesh]->mFaces[iAssFace].mIndices[2] + submeshRanges[iAssMesh].VertexStart;
			}
		}
	}
	wMesh->UnlockIndexBuffer();

	delete submeshRanges;

	struct Vertex{
		aiVector3D position;
		aiVector3D normal;
		aiVector3D tex;
	} *vertexData;
	wMesh->LockVertexBuffer(0, (void**)&vertexData);
	unsigned int iVertex=0;
	for(unsigned int iAssMesh=0; iAssMesh< assScene->mNumMeshes; iAssMesh++)
	{
		for(unsigned int iAssVertex=0; iAssVertex< assScene->mMeshes[iAssMesh]->mNumVertices; iAssVertex++)
		{
			vertexData[iVertex].position = assScene->mMeshes[iAssMesh]->mVertices[iAssVertex];
			vertexData[iVertex].normal = assScene->mMeshes[iAssMesh]->mNormals[iAssVertex];
			vertexData[iVertex].tex = assScene->mMeshes[iAssMesh]->mTextureCoords[0][iAssVertex];
			iVertex++;
		}
	}
	wMesh->UnlockVertexBuffer();

	if(wMesh == NULL)
	{
		MessageBoxA( NULL, 
			"Mesh could not be loaded.",
			"Failed to create mesh from file!", MB_OK);
		exit(-1);
	}

	std::vector<LPDIRECT3DTEXTURE9>::iterator iTexture = textures.begin();
	std::vector<LPDIRECT3DTEXTURE9>::iterator eTexture = textures.end();
	while(iTexture != eTexture )
	{
		if(*iTexture)
			(*iTexture)->Release();
		iTexture++;
	}
	textures.clear();
	textureFilenames.clear();

	char exePath[1024];

	GetModuleFileNameA( GetModuleHandle( L"x4egg.exe"), exePath, 1023);

	char* lastBackslash = strrchr(exePath, '\\');
	if(lastBackslash)
	{
		strcpy(lastBackslash, "/../media/uvGrid.jpg");
	}

	LPDIRECT3DTEXTURE9 defaultTexture;
	D3DXCreateTextureFromFileA( device, exePath, &defaultTexture);

	for(unsigned int iAssMesh=0; iAssMesh< assScene->mNumMeshes; iAssMesh++)
	{
		if(assScene->HasMaterials())
		{
			aiString texturePath;
			assScene->mMaterials[ assScene->mMeshes[iAssMesh]->mMaterialIndex ]->GetTexture(aiTextureType::aiTextureType_DIFFUSE, 0, &texturePath);
			LPDIRECT3DTEXTURE9 texture;
			HRESULT hr = D3DXCreateTextureFromFileA( device, texturePath.data, &texture);
			if(hr == S_OK)
			{
				textures.push_back(texture);
				textureFilenames.push_back(texturePath.data);
			}
			else
			{
				defaultTexture->AddRef();
				textures.push_back(defaultTexture);
				textureFilenames.push_back("media/uvGrid.jpg");
			}
		}
		else
		{
			defaultTexture->AddRef();
			textures.push_back(defaultTexture);
			textureFilenames.push_back("media/uvGrid.jpg");
		}
	}
	defaultTexture->Release();

	wMesh->OptimizeInplace(D3DXMESHOPT_ATTRSORT, NULL, NULL, NULL, NULL);

	mesh->Release();
	mesh = wMesh;

}

void App::saveFile()
{
	wchar_t filename[512];
	filename[0] = L'\0';
	OPENFILENAME of;
	of.lStructSize = sizeof(OPENFILENAME);
	of.hwndOwner = DXUTGetHWND();
	of.hInstance = NULL;
	of.lpstrFilter = L"D3D x file\0*.x\0\0";
	of.lpstrCustomFilter = NULL;
	of.nMaxCustFilter = 0;
	of.nFilterIndex = 0;
	of.lpstrFile = filename;
	of.nMaxFile = 512; 
	of.lpstrFileTitle = NULL;
	of.nMaxFileTitle = 0;
	of.lpstrInitialDir = L"."; 
	of.lpstrTitle = NULL;
	of.Flags = 0; 
	of.nFileOffset = 0; 
	of.nFileExtension = 0; 
	of.lpstrDefExt = L"x";
	of.lCustData = NULL;
	of.lpfnHook = NULL;
	of.lpTemplateName = NULL; 
	of.pvReserved = NULL;
	of.dwReserved = NULL;
	of.FlagsEx = 0;

	GetSaveFileName(&of);

	D3DXMATERIAL* materials = new D3DXMATERIAL[nSubMeshes];
	D3DXEFFECTINSTANCE* effects = new D3DXEFFECTINSTANCE[nSubMeshes];
	for(int i=0; i<nSubMeshes; i++)
	{
		materials[i].pTextureFilename = (LPSTR)textureFilenames.at(i).c_str();
		effects[i].NumDefaults = 0;
		effects[i].pEffectFilename = "nosuchfile.fx";
	}

	D3DXSaveMeshToX(filename, mesh, NULL, materials, effects, nSubMeshes, D3DXF_FILEFORMAT_TEXT);


	delete materials;
	delete effects;
}